/***************************** LICENSE START ***********************************

 Copyright 2012 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/

#pragma once

/************************************************************
  Application CptecAccess

  Access GRIB/BUFR data from CPTEC's storage directories
************************************************************/
#include <Metview.h>
//#include <sys/types.h>
//#include <sys/stat.h>

class CptecAccess : public MvService
{
public:
    CptecAccess(char* name) :
        MvService(name){};
    void serve(MvRequest&, MvRequest&);

    // Build the path
    bool buildPath(MvRequest&, std::string&);

    // Build the filenames
    bool buildFilenames(MvRequest&, std::vector<std::string>&);
};
