#Mass density (concentration)
'20000' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 0 ;
         centre = 65535;
        }

#Column-integrated mass density
'20001' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 1 ;
         centre = 65535;
        }

#Mass mixing ratio
'20002' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 2 ;
         centre = 65535;
        }

#Volume mixing ratio
'20052' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 52 ;
         centre = 65535;
        }
        
#FLEXPART tracer mass concentration
'20192' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 192 ;
         centre = 65535;
        }

#FLEXPART tracer mass ratio
'20193' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 193 ;
         centre = 65535;
        }

#FLEXPART particle resident time
'20194' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 194 ;
         centre = 65535;
        }

#FLEXPART response function for mass concentration
'20195' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 195 ;
         centre = 65535;
        }

#FLEXPART response function for mass ratio
'20196' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 196 ;
         centre = 65535;
        }

#FLEXPART dry deposition
'20197' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 197 ;
         centre = 65535;
        }

#FLEXPART wet deposition
'20198' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 198 ;
         centre = 65535;
        }

#FLEXPART eastward flux
'20199' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 199 ;
         centre = 65535;
        }

#FLEXPART westward flux
'20200' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 200 ;
         centre = 65535;
        }

#FLEXPART southward flux
'20201' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 201 ;
         centre = 65535;
        }

#FLEXPART northward flux
'20202' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 202 ;
         centre = 65535;
        }

#FLEXPART upward flux
'20203' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 203 ;
         centre = 65535;
        }

#FLEXPART downward flux
'20204' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 204 ;
         centre = 65535;
        }
