/***************************** LICENSE START ***********************************

 Copyright 2012 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/

#pragma once

#include <QMap>
#include <QPixmap>

class MvQPixmapCache
{
public:
    MvQPixmapCache(QPixmap);
    void add(QString key, QPixmap pix) { pixmaps_[key] = pix; }
    void addMvIconPath(QString, QString);
    QPixmap pixmap(QString);

private:
    QPixmap create(QString);
    QString mvIconPath(QString) const;

    QPixmap defaultPixmap_;
    int size_;
    QMap<QString, QPixmap> pixmaps_;
    QMap<QString, QString> mvIconPaths_;
};
