#Mass density (concentration)
'mdc' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 0 ;
        }

#Mass mixing ratio
'mmr' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 2 ;
        }

#FLEXPART tracer mass concentration
'ftmc' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 192 ;
        }

#FLEXPART tracer mass ratio
'ftmr' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 193 ;
        }

#FLEXPART particle resident time
'fprt' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 194 ;
        }

#FLEXPART response function for mass concentration
'frfmc' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 195 ;
        }

#FLEXPART response function for mass ratio
'frfmr' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 196 ;
        }

#FLEXPART dry deposition concentration
'fddc' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 197 ;
        }

#FLEXPART wet deposition concentration
'fwdc' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 198 ;
        }

#FLEXPART dry deposition mixing ratio
'fddmr' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 199 ;
        }

#FLEXPART wet deposition mixing ratio
'fwdmr' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 200 ;
       }

#FLEXPART eastward flux
'feflux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 201 ;
        }

#FLEXPART westward flux
'fwflux' = {
          discipline = 0 ;
          parameterCategory = 20 ;
          parameterNumber = 202 ;
         }

#FLEXPART southward flux
'fsflux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 203 ;
        }

#FLEXPART northward flux
'fnflux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 204 ;
        }

#FLEXPART upward flux
'fuflux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 205 ;
        }

#FLEXPART downward flux
'fdflux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 206 ;
        }
