1 ; HIRS ; HIRS
2 ; AMSUA ; AMSUA
3 ; AMSUB ; AMSUB
4 ; MHS ; MHS
5 ; GEOS ; GEOS
6 ; RESAT ; RESAT
7 ; MERIS ; MERIS
8 ; GPSRO ; GPSRO
9 ; SATOB ; SATOB
10 ; SCATT ; SCATT
11 ; SSMI All-sky ; SSMI_AS
12 ; IASI ; IASI
13 ; AIRS ; AIRS
14 ; SSMIS All-sky ; SSMIS_AS
15 ; TMI All-sky ; TMI_AS
16 ; AMSRE All-sky ; AMSRE_AS
17 ; CONV ; CONV
19 ; SMOS ; SMOS
20 ; WINDSAT All-sky ; WINDSAT_AS
21 ; SSMI ; SSMI
22 ; AMSUA All-sky ; AMSUA_AS
23 ; AMSRE ; AMSRE
24 ; TMI ; TMI
25 ; SSMIS ; SSMIS
26 ; GBRAD ; GBRAD
27 ; MWHS ; MWHS
28 ; MWTS ; MWTS
29 ; MWRI All-sky ; MWRI_AS
30 ; IRAS ; IRAS
31 ; MSU ; MSU
32 ; SSU ; SSU
33 ; VTPR1 ; VTPR1
34 ; VTPR2 ; VTPR2
35 ; ATMS ; ATMS
36 ; RESAT Averaging Kernels ; RESAT_AK
37 ; CRIS ; CRIS
38 ; WAVE integrated Parameters ; WAVE_IP
39 ; WAVE spectra ; WAVE_SP
40 ; RAINGG ; RAINGG
41 ; SURFACE MULTISENSOR ; SFC_MS
42 ; AMSR-2 All-sky ; AMSR2_AS
43 ; SAPHIR All-sky ; SAPHIR_AS
44 ; AMSUB All-sky ; AMSUB_AS
45 ; MHS All-sky ; MHS_AS
46 ; Doppler Wind Lidar ; DWL
47 ; IRIS ; IRIS
49 ; AATSR ; AATSR
50 ; ATMS All-sky ; ATMS_AS
51 ; GMI All-sky ; GMI_AS
52 ; GODAE Sea Surface Temperatures ; GODAE_SST
53 ; ATOVS MULTISENSOR ; ATOVS_MS
54 ; Atmospheric composition ; ATMOSPHERIC_COMPOSITION
55 ; NON-SURFACE MULTISENSOR ; NON_SFC_MS
56 ; MWTS2 ; MWTS2
57 ; SSMI 1DVAR TCWV Cloudy-Sky ; SSMI_1D
58 ; MWHS2 All-sky ; MWHS2_AS
59 ; SSMT2 ; SSMT2
60 ; SMAP ; SMAP
61 ; TOVS MULTISENSOR ; TOVS_MS
62 ; Cloud reflectivity ; CLOUD_R
63 ; Cloud Lidar ; CLOUD_L
66 ; OCONV ; OCONV
99 ; TEST ; TEST
