/***************************** LICENSE START ***********************************

 Copyright 2012 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/

#pragma once

#include "Metview.h"

#include <QList>
#include <QObject>

class QAction;

class Tools : public QObject
{
    Q_OBJECT

public:
    Tools(request*);
    ~Tools();

    static void make(QList<QAction*>&, QObject*);
    static void load(request*);

public slots:
    void start();

private:
    // No copy allowed
    Tools(const Tools&);
    Tools& operator=(const Tools&);

    request* request_;
};

inline void destroy(Tools**) {}
