unsigned[2] flexpartVersion  : dump ;
unsigned[1] speciesId : dump ;
ascii[10] speciesName : dump ;
unsigned[1] numberOfReleases : dump ;
unsigned[1] releaseNumber : dump ;
ascii[40] releaseName : dump ;
unsigned[1] numberOfAgeClasses : dump ;
unsigned[1] ageClass : dump ;
unsigned[4] ageClassBegin : dump ;
unsigned[4] ageClassEnd : dump ;
