/***************************** LICENSE START ***********************************

 Copyright 2018 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/

#pragma once

#include <vector>
#include "MvGeoPoints.h"

// currently a dummy class, maybe to be used later


class MvGeoPointSet : public std::vector<MvGeoPoints*>
{
public:
    // constructor
    MvGeoPointSet();
    ~MvGeoPointSet();
};
