#Mass density (concentration)
'Mass density (concentration)' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 0 ;
         centre = 65535;
        }

#Column-integrated mass density
'Column-integrated mass density' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 1 ;
         centre = 65535;
        }

#Mass mixing ratio
'Mass mixing ratio' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 2 ;
         centre = 65535;
        }

#Volume mixing ratio
'Volume mixing ratio' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 52 ;
         centre = 65535;
        }

#FLEXPART tracer mass concentration
'FLEXPART tracer mass concentration' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 192 ;
         centre = 65535;
        }

#FLEXPART tracer mass ratio
'FLEXPART tracer mass ratio' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 193 ;
         centre = 65535;
        }

#FLEXPART particle resident time
'FLEXPART particle resident time' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 194 ;
         centre = 65535;
        }

#FLEXPART response function for mass concentration
'FLEXPART response function for mass concentration' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 195 ;
         centre = 65535;
        }

#FLEXPART response function for mass ratio
'FLEXPART response function for mass ratio' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 196 ;
         centre = 65535;
        }

#FLEXPART dry deposition
'FLEXPART dry deposition' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 197 ;
         centre = 65535;
        }

#FLEXPART wet deposition
'FLEXPART wet deposition' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 198 ;
         centre = 65535;
        }

#FLEXPART eastward flux
'FLEXPART eastward flux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 199 ;
         centre = 65535;
        }

#FLEXPART westward flux
'FLEXPART westward flux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 200 ;
         centre = 65535;
        }

#FLEXPART southward flux
'FLEXPART southward flux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 201 ;
         centre = 65535;
        }

#FLEXPART northward flux
'FLEXPART northward flux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 202 ;
         centre = 65535;
        }

#FLEXPART upward flux
'FLEXPART upward flux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 203 ;
         centre = 65535;
        }

#FLEXPART downward flux
'FLEXPART downward flux' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 204 ;
         centre = 65535;
        }
