#Mass density (concentration)
'20000' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 0 ;
        }

#Mass mixing ratio
'20002' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 2 ;
        }

#FLEXPART tracer mass concentration
'20192' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 192 ;
        }

#FLEXPART tracer mass ratio
'20193' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 193 ;
        }

#FLEXPART particle resident time
'20194' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 194 ;
        }

#FLEXPART response function for mass concentration
'20195' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 195 ;
        }

#FLEXPART response function for mass ratio
'20196' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 196 ;
        }

#FLEXPART dry deposition concentration
'20197' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 197 ;
        }

#FLEXPART wet deposition concentration
'20198' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 198 ;
        }

#FLEXPART dry deposition mixing ratio
'20199' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 199 ;
        }

#FLEXPART wet deposition mixing ratio
'20200' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 200 ;
        }

#FLEXPART eastward flux
'20201' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 201 ;
        }

#FLEXPART westward flux
'20202' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 202 ;
        }

#FLEXPART southward flux
'20203' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 203 ;
        }

#FLEXPART northward flux
'20204' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 204 ;
        }

#FLEXPART upward flux
'20205' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 205 ;
        }

#FLEXPART downward flux
'20206' = {
         discipline = 0 ;
         parameterCategory = 20 ;
         parameterNumber = 206 ;
        }
