/*
 * © Copyright 1996-2012 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

typedef struct ecaccess_info {
    /* char  *echost; */
    /* char  *ecaccesshome; */
    /* char  *ectool; */
    /* int   eccertport; */
    char connect_host[80];
    int connect_port;
    char uid[80];
} ecaccess_info;
